#
#   Tiles placer and layers renderer.
#   Copyright (C) 2020  Alessandro Accattini
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#*****************************************************************************
#
#
#


## \brief MapInfo contains general scene settings and tiles dimensions.
#
set MapInfo(OffX) 300			;# First tile image offset
set MapInfo(OffY) 30
set MapInfo(TileW) 37			;# Tile dimensions
set MapInfo(TileH) 17
set MapInfo(Cols) 0			;# Map dimensions
set MapInfo(Rows) 0
set MapInfo(Scans) 0			;# Map total scan lines

## \brief Directories info struct, contains all data source directories.
#
set Dir(Tiles) .			;# tiles source directory
set Dir(Maps) .				;# maps and layers source directory.

## \brief Tiles database.
#
#  Database is a list of lists, one for each Db file line.
#
set TilesDb {}

## \brief Layers database.
#
#  Database is a list of layers heights.
#
set LayersDb {}

## \brief Current layer map.
#
set Map {}





## \brief Configures source directories.
#  \param TilesPath         Tiles source directory.
#  \param MapsPath          Maps and layers source directory.
#
proc SetSourceDir {TilesPath MapsPath} {
  global Dir

  set Dir(Tiles) $TilesPath
  set Dir(Maps) $MapsPath
}



## \brief Configures scene parameters.
#  \param FilrstTileOffX    First map tile horizontal offset in the scene.
#  \param FilrstTileOffY    First map tile vertical offset in the scene.
#  \praam TileSizeX         Tiles width.
#  \praam TileSizeY         Tiles heigth.
#
proc SetScene {FirstTileOffX FirstTileOffY TileSizeX TileSizeY} {
  global MapInfo

  set MapInfo(OffX) $FirstTileOffX
  set MapInfo(OffY) $FirstTileOffY
  set MapInfo(TileW) $TileSizeX
  set MapInfo(TileH) $TileSizeY
}



## \brief Read tiles database file.
#  \param TilesDbFileName   Database file name.
#  \return                  Nothing.
#
proc LoadTilesDb {TilesDbFileName} {
  global Dir
  global TilesDb

  set TilesDb {}				;# empty DB
  set FullFileName $Dir(Maps)/$TilesDbFileName	;# creates full file name
  puts "Loading: $FullFileName"
  set File [open $FullFileName r]		;# open file to read
  while {[gets $File DbLine] >= 0} {		;# read a line from file
    lappend TilesDb [split $DbLine ,]		;# convert line to list
  }
}



## \brief Read layers database file.
#  \param LayersDbFileName   Database file name.
#  \return                   Nothing.
#
proc LoadLayersDb {LayersDbFileName} {
  global Dir
  global LayersDb

  set LayersDb {}				;# empty DB
  set FullFileName $Dir(Maps)/$LayersDbFileName	;# creates full file name
  puts "Loading: $FullFileName"
  set File [open $FullFileName r]		;# open file to read
  while {[gets $File DbLine] >= 0} {		;# read a line from file
    lappend LayersDb $DbLine			;# add line to DB
  }
}



## \brief Load current layer map.
#  \param MapFileName        Map file name.
#  \return                   Nothing.
#
proc LoadMap {MapFileName} {
  global Dir
  global Map
  global MapInfo

  set Map {}					;# starts from empty map
  set FullFileName $Dir(Maps)/$MapFileName	;# creates full file name
  puts "Loading: $FullFileName"
  set File [open $FullFileName r]		;# open map file to read
  while {[gets $File MapLine] >= 0} {		;# read a line from map file
    lappend Map [split $MapLine ,]		;# convert line to list
  }
  set MapInfo(Rows) [llength $Map]
  set MapInfo(Cols) [llength [lindex $Map 1]]
  set MapInfo(Scans) [expr $MapInfo(Cols) + $MapInfo(Cols) - 1]
}



## \brief Adds a single tile to current scene image file.
#  \param SceneFileName    Current scene file name.
#  \param MapX             Map coordinate X.
#  \param MapY             Map coordinate Y.
#  \param Layer            Map layer number (starting from 0).
#  \param TileName         Tile ID in tiles database.
#  \return                 Nothing.
#
proc AddTile {SceneFileName MapX MapY Layer TileName} {
  global Dir

  ;# checks for special tile names
  if {$TileName == "nul"} {
    return;					;# ignore it
  }
  set Info [TileInfo $TileName]			;# get tile info
  set FileName [lindex $Info 1]			;# extracts tile file name
  set TileFile $Dir(Tiles)/$FileName		;# creates full file name
  set OffX [lindex $Info 2]			;# extracts tile offsets
  set OffY [lindex $Info 3]
  set OffV [VerticalOffset $Layer]		;# vertical offset
  set OffY [expr $OffY + $OffV]			;# adjust vertical position
  set Geo [MapToSceneGeometry $MapX $MapY $OffX $OffY]

  ;# add tile to scene
  exec -- composite -geometry $Geo $TileFile $SceneFileName $SceneFileName
}



## \brief Adds a full layer to the current scene.
#  \param SceneFileName    Current scene file name.
#  \param LayerMapFileName Map file name.
#  \param Layer            Map layer number (starting from 0).
#
proc AddLayer {SceneFileName LayerMapFileName Layer} {
  global MapInfo

  LoadMap $LayerMapFileName			;# loads layer map
  for {set i 0} {$i < $MapInfo(Scans)} {incr i} {
    set Scan [GetScanLine $i]			;# creates scan line list
    for {set j 0} {$j < [llength $Scan]} {incr j} {
      set Coord [lindex $Scan $j]		;# pick coord. pair from list
      set x [lindex $Coord 0]			;# extracs coordinates
      set y [lindex $Coord 1]

      puts -nonewline "."
      flush stdout
      AddTile $SceneFileName $x $y $Layer [GetTileName $x $y]
    }
  }
  puts ""
}



## \brief Calculates vertical offset for current layer.
#  \param Layer       Active layer.
#  \return            Vertical offset.
proc VerticalOffset {Layer} {
  global LayersDb

  set Voff 0
  while {$Layer > 0} {
    incr Layer -1				;# back one position
    set Voff [expr $Voff + [lindex $LayersDb $Layer]]
  }
  return $Voff
}



## \brief Retrieves tile info from database.
#  \param TileName    Tile name ID.
#  \return            Tile data.
#
proc TileInfo {TileName} {
  global TilesDb

#  set Idx [lsearch -regexp $TilesDb $TileName*]
  for {set i 0} {$i < [llength $TilesDb]} {incr i} {
    set Entry [lindex $TilesDb $i]
    set ID [lindex $Entry 0]
    if {$ID == $TileName} {
      return $Entry				;# found
    }
  }
#  puts "Found at: $Idx"
#  return [lindex $TilesDb $Idx]
  return {}					;# not found
}



## \brief Converts map coordinates in to scene coordinates.
#  \param MapX             Map coordinate X.
#  \param MapY             Map coordinate Y.
#  \param BaseOffX         Tile base center offset X.
#  \param BaseOffY         Tile base center offset Y.
#  \returns                Tile geometry.
#
proc MapToSceneGeometry {MapX MapY BaseOffX BaseOffY} {
  global MapInfo

  set x [expr $MapInfo(OffX) - $BaseOffX - ($MapInfo(TileW) * ($MapY - $MapX))]
  set y [expr $MapInfo(OffY) - $BaseOffY + ($MapInfo(TileH) * ($MapY + $MapX))]
  if {$x >= 0} {
    set x "+$x"
  }
  if {$y >= 0} {
    set y "+$y"
  }
  return "$x$y"
}



## \breif Creates a diagonal scan line
#  \param ScanIndex[in]      Diagonal line index (stating from 0).
#  return                    A list of scan line map coordinates.
#
proc GetScanLine {ScanIndex} {
  global MapInfo
  
  if {$ScanIndex >= $MapInfo(Scans)} {		;# if out of the map
    return {}					;# return empty list
  }
  set ScanLine {}				;# starts from empty scan line
  if {$ScanIndex <= [expr $MapInfo(Cols) - 1]} {
    ;# you are on the upper left triangle of the map
    set Nelem [expr $ScanIndex + 1]
    for {set i 0} {$i < $Nelem} {incr i} {
      set y [expr $ScanIndex - $i]
      set x $i
      set Entry {}				;# empty list
      lappend Entry $x
      lappend Entry $y
      lappend ScanLine $Entry			;# add one tile coordinates
    }
    return $ScanLine
  }
  
  ;# you are on the lower right triangle of the map
  set Nelem [expr ($MapInfo(Cols) - 1) - ($ScanIndex - $MapInfo(Rows))]
  for {set i 0} {$i < $Nelem} {incr i} {
    set y [expr ($MapInfo(Rows) - 1) - $i]
    set x [expr ($ScanIndex - $MapInfo(Rows)) + 1 + $i]
    set Entry {}				;# empty list
    lappend Entry $x
    lappend Entry $y
    lappend ScanLine $Entry			;# add one tile coordinates
  }
  return $ScanLine
}



## \brief Gets tile name from map coordinates.
#  \param MapX             Map coordinate X.
#  \param MapY             Map coordinate Y.
#  \return                 Tile ID.
#
proc GetTileName {MapX MapY} {
  global Map

  return [lindex $Map $MapY $MapX]
}
